#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <memory>
#include <iostream>

#define Uses_EnhancedMatrix
#include "matrix"

//#define KECY

using namespace std;

CEnhancedMatrix::CEnhancedMatrix()
  : CMatrix()
{
#ifdef KECY
  cout << "CEnhancedMatrix::CEnhancedMatrix();" << endl;
#endif
  m_blocks = n_blocks = 0; BlockMatrix = NULL;
}
CEnhancedMatrix::CEnhancedMatrix(int _m, int _n)
  : CMatrix(_m, _n)
{
#ifdef KECY
  cout << "CEnhancedMatrix::CEnhancedMatrix(" << _m << ", " << _n << ");" << endl;
#endif
  m_blocks = n_blocks = 0; BlockMatrix = NULL;
}
CEnhancedMatrix::CEnhancedMatrix(const CMatrix& Matrix)
  : CMatrix(Matrix)
{
#ifdef KECY
  cout << "CEnhancedMatrix::CEnhancedMatrix(Matrix);" << endl;
#endif
  m_blocks = n_blocks = 0; BlockMatrix = NULL;
}
CEnhancedMatrix::CEnhancedMatrix(int _m, int _n, char *_matrix)
  : CMatrix(_m, _n, _matrix)
{
#ifdef KECY
  cout << "CEnhancedMatrix::CEnhancedMatrix(" << _m << ", " << _n << ", \"_matrix\");" << endl;
#endif
  m_blocks = n_blocks = 0; BlockMatrix = NULL;
}

CEnhancedMatrix::~CEnhancedMatrix()
{
#ifdef KECY
  cout << "CEnhancedMatrix::~CEnhancedMatrix();" << endl;
#endif
}

void CEnhancedMatrix::SetupBlockMatrix(int m0, ...)
{
  va_list arglist;
  int m_count = 1 /* add the initial one */, n_count = 0;
  int m_mode, n_mode;
  int n0;
  int *m_list = NULL, *n_list = NULL;
  int i;

  /* Destroy the existing block matrix */
  DestroyBlockMatrix();

  /* get count of parameters */
  va_start(arglist, m0);
  while ((m_mode = va_arg(arglist, int)) > 0) m_count++;
  while ((n_mode = va_arg(arglist, int)) > 0) n_count++;
  va_end(arglist);

  va_start(arglist, m0);

  try {
    switch (m_mode)
    {
      case BM_USERDEFINED: /* user-defined rows */
        try {
          m_list = new int[m_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (m_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);

        m_list[0] = m0;
        for (i = 1; i < m_count; i++)
          m_list[i] = va_arg(arglist, int);
        (void)va_arg(arglist, int); /* skip mode selector */
        break;
        
      case BM_SIZE:        /* automatically created rows with given height */
        m_count = (GetRows() + m0 - 1) / m0; /* compute count of rows */
        try {
          m_list = new int[m_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (m_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);

        for (i = 0; i < m_count; i++)
          m_list[i] = m0;
        if (m_count * m0 > GetRows()) /* eliminate overflow */
          m_list[m_count - 1] -= m_count * m0 - GetRows();
        (void)va_arg(arglist, int); /* skip mode selector */
        break;

      case BM_COUNT:      /* automatically created specified amount of rows */
        m_count = m0;
        try {
          m_list = new int[m_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (m_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);
        
        for (i = 0; i < m_count; i++)
          m_list[i] = GetRows() / m_count; /* compute row height */
        if (m_list[0] * m_count < GetRows()) /* eliminate underflow */
          m_list[0] += GetRows() - m_list[0] * m_count;
        (void)va_arg(arglist, int); /* skip mode selector */
        break;
    }

    switch (n_mode)
    {
      case BM_USERDEFINED: /* user-defined columns */
        try {
          n_list = new int[n_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (n_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);
        
        for (i = 0; i < n_count; i++)
          n_list[i] = va_arg(arglist, int);

        (void)va_arg(arglist, int); /* skip mode selector */
        break;

      case BM_SIZE:        /* automatically created columns with given width */
        n0 = va_arg(arglist, int);
        n_count = (GetCols() + n0 - 1) / n0; /* compute count of rows */
        if (n_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);
        
        try {
          n_list = new int[n_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (n_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);
          
        for (i = 0; i < n_count; i++)
          n_list[i] = n0;
        if (n_count * n0 > GetCols()) /* eliminate overflow */
          n_list[n_count - 1] -= n_count * n0 - GetCols();
        (void)va_arg(arglist, int); /* skip mode selector */
        break;

      case BM_COUNT:      /* automatically created specified amount of columns */
        n_count = va_arg(arglist, int);

        try {
          n_list = new int[n_count];
        } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
        if (n_list == NULL)
          throw CMatrixError(M_NO_MEM_AUX);
        
        for (i = 0; i < n_count; i++)
          n_list[i] = GetCols() / n_count; /* compute row height */
        if (n_list[0] * n_count < GetCols()) /* eliminate underflow */
          n_list[0] += GetCols() - n_list[0] * n_count;
        (void)va_arg(arglist, int); /* skip mode selector */
        break;
    }
  } catch (CMatrixError)
  {
    if (n_list != NULL)
      delete[] n_list;
    if (m_list != NULL)
      delete[] m_list;
    throw;
  }
  
  va_end(arglist);

  SetupCustomBlockMatrix(m_count, m_list, n_count, n_list);

  delete[] n_list;
  delete[] m_list;
}

void CEnhancedMatrix::SetupCustomBlockMatrix(int _m, int *m_list, int _n, int *n_list)
{
  int m_entry, n_entry;
  int i, j, k, l;

  m_blocks = _m;
  n_blocks = _n;

  /* create new matrix blocks */
  try {
    BlockMatrix = new CEnhancedMatrix*[m_blocks * n_blocks];
  } catch(bad_alloc) { throw CMatrixError(M_NO_MEM); }
  /* not enough memory? */
  if (BlockMatrix == NULL)
    throw CMatrixError(M_NO_MEM);

  m_entry = 0;
  for (i = 0; i < m_blocks; i++)
  {
    n_entry = 0;
    for (j = 0; j < n_blocks; j++)
    {
      /* creare matrix block */
      try {
        BlockMatrix[i * n_blocks + j] = new CEnhancedMatrix(m_list[i], n_list[j]);
      } catch(bad_alloc) { throw CMatrixError(M_NO_MEM); }
      if (BlockMatrix[i * n_blocks + j] == NULL)
        throw CMatrixError(M_NO_MEM);

      /* copy numbers for original matrix to the new block */
      for (k = 0; k < m_list[i]; k++)
        for (l = 0; l < n_list[j]; l++)
          BlockMatrix[i * n_blocks + j]->SetField(k, l,
            GetField(m_entry + k, n_entry + l));

      n_entry += n_list[j];
    }
    m_entry += m_list[i];
  }
}

void CEnhancedMatrix::DestroyBlockMatrix()
{
  int i;
  if (BlockMatrix != NULL)
  {
    for (i = 0; i < m_blocks * n_blocks; i++)
      if (BlockMatrix[i] != NULL)
        delete BlockMatrix[i];
    delete[] BlockMatrix;
  }
  m_blocks = n_blocks = 0;
  BlockMatrix = NULL;
}

/* Reflects changes to the block matrix back to the original matrix */
void CEnhancedMatrix::UpdateMatrix()
{
  int m_entry, n_entry;
  int i, j, k, l;
  int _m, _n;

  m_entry = 0;
  for (i = 0; i < m_blocks; i++)
  {
    n_entry = 0;
    for (j = 0; j < n_blocks; j++)
    {
      BlockMatrix[i * n_blocks + j]->GetSize(_m, _n);

      /* copy numbers for original matrix to the new block */
      for (k = 0; k < _m; k++)
        for (l = 0; l < _n; l++)
          SetField(m_entry + k, n_entry + l,
            BlockMatrix[i * n_blocks + j]->GetField(k, l));

      n_entry += _n;
    }
    m_entry += _m;
  }
}

/* Gets block matrix dimensions. */
void CEnhancedMatrix::GetBlocksSize(int &_m, int &_n)
{
  _m = m_blocks;
  _n = n_blocks;
}
int CEnhancedMatrix::GetBlocksRows()
{
  return m_blocks;
}
int CEnhancedMatrix::GetBlocksCols()
{
  return n_blocks;
}

CEnhancedMatrix& CEnhancedMatrix::GetBlockMatrix(int _m, int _n)
{
  if (BlockMatrix != NULL)
  {
    if ((_m < 0) || (_m >= m_blocks) || (_n < 0) || (_n >= n_blocks))
      throw CMatrixError(M_OUT_OF_BOUNDS);
    return *BlockMatrix[_m * n_blocks + _n];
  }
  throw CMatrixError(M_NO_BLKMTX);
}

void CEnhancedMatrix::ExchangeBlockRows(int _m1, int _m2)
{
  if (BlockMatrix == NULL)
    throw CMatrixError(M_NO_BLKMTX);

  CEnhancedMatrix **tmp;
  try {
    tmp = new CEnhancedMatrix*[n_blocks];
  } catch(bad_alloc) { throw CMatrixError(M_NO_MEM_AUX); }
  if (tmp == NULL)
    throw CMatrixError(M_NO_MEM_AUX);

  memcpy(tmp, &BlockMatrix[_m1 * n_blocks], sizeof(CEnhancedMatrix *) * n_blocks);
  memcpy(&BlockMatrix[_m1 * n_blocks], &BlockMatrix[_m2 * n_blocks], sizeof(CEnhancedMatrix *) * n_blocks);
  memcpy(&BlockMatrix[_m2 * n_blocks], tmp, sizeof(CEnhancedMatrix *) * n_blocks);

  delete[] tmp;
}

