#ifndef __ENHANCEDMATRIX_H
#define __ENHANCEDMATRIX_H

#define BM_USERDEFINED 0
#define BM_SIZE        -1
#define BM_COUNT       -2

class CEnhancedMatrix : public CMatrix
{
public:
  CEnhancedMatrix();
  CEnhancedMatrix(int _m, int _n);
  CEnhancedMatrix(const CMatrix& Matrix);
  CEnhancedMatrix(int _m, int _n, char *_matrix);

  virtual ~CEnhancedMatrix();

  /* Makes a block matrix from ordinary matrix (just splits the numbers into
     blocks).
     This function has rather complex syntax:

       How to create user defined block matrix with exactly specified
       block sizes:
         SetupBlockMatrix(mblocksize_0, ..., mblocksize_m, 0,
           nblocksize_0, ..., nblocksize_n, 0);
       How to split matrix into rows with height row_m and columns with
       width col_n:
         SetupBlockMatrix(row_m_0, -1, col_n, -1);
       How to split matrix into m rows and n columns:
         SetupBlockMatrix(m, -2, n, -2);

     This function is only a front-end for SetupCustomBlockMatrix
  */
  void SetupBlockMatrix(int m0, ...);
  void SetupCustomBlockMatrix(int _m, int *m_list, int _n, int *n_list);
  void DestroyBlockMatrix();

  /* Reflects changes to the block matrix back to the original matrix */
  void UpdateMatrix();

  /* Gets block matrix dimensions. */
  void GetBlocksSize(int &_m, int &_n);
  int GetBlocksRows();
  int GetBlocksCols();

  /* Gets matrix block */
  CEnhancedMatrix& GetBlockMatrix(int _m, int _n);

  /* Exchanges two block rows */
  void ExchangeBlockRows(int _m1, int _m2);
  
private:  
  int m_blocks, n_blocks;
  CEnhancedMatrix **BlockMatrix;
};

#endif
