#include <cstdio>
#include <iostream>

#define Uses_EnhancedMatrix
#define Uses_GaussMatrix
#define Uses_PrintMatrix
#include "matrix"

//#define KECY

using namespace std;

CGaussMatrix::CGaussMatrix() :
  CEnhancedMatrix()
{
#ifdef KECY
  cout << "CGaussMatrix::CGaussMatrix();" << endl;
#endif
}

CGaussMatrix::CGaussMatrix(int _m, int _n) :
  CEnhancedMatrix(_m, _n)
{
#ifdef KECY
  cout << "CGaussMatrix::CGaussMatrix(" << _m << ", " << _n << ");" << endl;
#endif
}

CGaussMatrix::CGaussMatrix(const CMatrix& Matrix) :
  CEnhancedMatrix(Matrix)
{
#ifdef KECY
  cout << "CGaussMatrix::CGaussMatrix(Matrix);" << endl;
#endif
}

CGaussMatrix::CGaussMatrix(int _m, int _n, char *_matrix) :
  CEnhancedMatrix(_m, _n, _matrix)
{
#ifdef KECY
  cout << "CGaussMatrix::CGaussMatrix(" << _m << ", " << _n << ", \"_matrix\");" << endl;
#endif
}

CGaussMatrix::~CGaussMatrix()
{
#ifdef KECY
  cout << "CGaussMatrix::~CGaussMatrix();" << endl;
#endif
}

void CGaussMatrix::Eliminate(int columns)
{
  int i, j;
  /* int k; */
  CEnhancedMatrix tmp;

  /* avoid overflow */
  if (columns > GetCols())
    throw CMatrixError(M_OUT_OF_BOUNDS);

  /* Split matrix into row into block matrices */
  SetupBlockMatrix(1, BM_SIZE, 1, BM_COUNT);

#ifdef KECY
  cout << "Pass 1" << endl;
#endif

  /* Pass 1 */
  for (i = 0; i < columns; i++)
  {
    /* let all rows have 1 or 0 on the field 'i' */
    for (j = i; j < GetRows(); j++)
    {
      /* if the row 'j' has 0 on the field 'i', move it down, if possible */
      if ((GetBlockMatrix(j, 0).GetField(0, i) == 0)
        && (j < GetRows() - 1))
        ExchangeBlockRows(j, j + 1);	
      if (GetBlockMatrix(j, 0).GetField(0, i) != 0)
        GetBlockMatrix(j, 0) *= 1.0 / GetBlockMatrix(j, 0).GetField(0, i);
    }
    
#ifdef KECY
    cout << "Status in iteration " << i << ":" << endl;
    for (j = 0; j < GetRows(); j++)
      PrintMatrix(GetBlockMatrix(j, 0));
#endif

    /* substract row 'i' from all other rows */
    for (j = i+1; j < GetRows(); j++)
    {
      /* substract only from rows which do not have zero field 'i' */
      if (GetBlockMatrix(j, 0).GetField(0, i) != 0)
        GetBlockMatrix(j, 0) -= GetBlockMatrix(i, 0);
    }

#ifdef KECY
    cout << "Status after iteration " << i << ":" << endl;
    for (j = 0; j < GetRows(); j++)
      PrintMatrix(GetBlockMatrix(j, 0));
#endif
  }

#ifdef KECY
  cout << "Pass 2" << endl;
#endif

  /* Pass 2 */
  /* avoid overflow */
  int maxrows = (columns > GetRows() ? GetRows() : columns);
  
  for (i = maxrows - 1; i >= 0; i--)
  {
//    k = 0;
//    while ((i - k > 0) && (GetBlockMatrix(i - k, 0).GetField(0, i) == 0)) k++;
//    if (i - k > 0) /* prevent singularities */
    if (GetBlockMatrix(i, 0).GetField(0, i) != 0) /* prevent singularities */
      /* substract row 'i' from all other rows */
      for (j = i /*- k*/ - 1; j >= 0; j--)
      {
        tmp = GetBlockMatrix(i /*- k*/, 0) *
          (GetBlockMatrix(j, 0).GetField(0, i) / GetBlockMatrix(i /*- k*/, 0).GetField(0, i));
        GetBlockMatrix(j, 0) -= tmp;
      }

#ifdef KECY
    cout << "Status after iteration " << i << ":" << endl;
    for (j = 0; j < GetRows(); j++)
      PrintMatrix(GetBlockMatrix(j, 0));
#endif
  }

  /* Update the original matrix */
  UpdateMatrix();

  DestroyBlockMatrix();
}

void CGaussMatrix::LinearEquation()
{
  Eliminate(GetCols() - 1);
}
