#ifndef __MATRIX_H
#define __MATRIX_H

/* invalid matrix size (must be at least 1x1) */
#define M_INV_SIZE      1
/* field address is out of bounds */
#define M_OUT_OF_BOUNDS 2
/* matrix has bad size */
#define M_BAD_SIZE      3
/* not enough memory */
#define M_NO_MEM        4
/* not enough memory for auxiliary operation */
#define M_NO_MEM_AUX    5
/* block matrix not initialized */
#define M_NO_BLKMTX     6

#define M_MAX_ERROR     6

//#define KECY

class CMatrix
{
public:
  CMatrix();
  CMatrix(int _m, int _n);
  CMatrix(const CMatrix& Matrix);
  CMatrix(int _m, int _n, char *_matrix);
  virtual ~CMatrix();

  /* Fills the matrix using real numbers.
     WARNING: You have to specify the numbers EXACTLY this way:
       Setup(1.0, 2.0, 3.0)
     If you use Setup(1, 2, 3), all numbers except the first one will be
     corrupted. If you need to pass integers, use SetupInt instead. */
  void Setup(double d, ...);

  /* Fills the matrix using integers. */
  void SetupInt(long int d, ...);
  
  /* Fills the matrix using a string. */
  void SetupStr(char *_matrix);

  /* Gets matrix field m,n. */
  virtual double GetField(int i, int j) const;

  /* Sets matrix field m,n */
  virtual void SetField(int i, int j, double d);
  
  /* Gets matrix dimensions. */
  void GetSize(int &_m, int &_n) const;

  virtual int GetCols();
  virtual int GetRows();

  /* Gets matrix value */
  double *GetMatrix();
  double *GetMatrix(int &_m, int &_n);

  CMatrix& operator =  (const CMatrix& rvalue);
  CMatrix& operator += (const CMatrix& rvalue);
  CMatrix& operator -= (CMatrix& rvalue);
  CMatrix& operator *= (double rvalue);
  CMatrix& operator *= (CMatrix& rvalue);

  /* return this matrix as transposed */
  CMatrix MakeTransposed();

  /* transpose this matrix */
  void Transpose();

  /* matrix addition */
  friend CMatrix operator + (const CMatrix& left, const CMatrix& right);

  /* matrix multiplication (scalar) */
  friend CMatrix operator * (const CMatrix& left, double right);
  friend CMatrix operator * (double left, const CMatrix& right);

  /* matrix negation */
  friend CMatrix operator - (const CMatrix& right);

  /* matrix substraction */
  friend CMatrix operator - (const CMatrix& left, const CMatrix& right);

  /* matrix multiplication */
  friend CMatrix operator * (const CMatrix& left, const CMatrix& right);

  /* matrix comparison */
  friend int operator == (const CMatrix& left, const CMatrix& right);
  friend int operator != (const CMatrix& left, const CMatrix& right);

private:
  int m, n;
  double *matrix;
  
  /* These creation functions throw exceptions, so nothing should be
     allocated before calling them. */
  /* Creates a new matrix (m) x (n) (m rows and n columns) */
  void CreateMatrix(int _m, int _n);
  /* Creates a new matrix from another matrix */
  void CreateMatrix(const CMatrix& Matrix);

  /* Destroys the matrix */
  void DestroyMatrix();
};

class CMatrixError: public std::exception
{
public:
  CMatrixError(int _id);
  virtual ~CMatrixError() throw();
  int ID() const;
  virtual const char* what() const throw();

  CMatrixError& operator = (const CMatrixError& rvalue);

  friend int operator == (const CMatrixError& left, const CMatrixError& right);

private:
  int id;
};

#endif /* __MATRIX_H */
